#include <stdio.h>

#include <osmocom/core/select.h>
#include <osmocom/core/utils.h>
#include <osmocom/core/logging.h>
#include <osmocom/core/application.h>
#include <osmocom/core/fsm.h>
#include <osmocom/vty/vty.h>
#include <osmocom/vty/telnet_interface.h>

#include <osmocom/sigtran/osmo_ss7.h>
#include <osmocom/sigtran/sccp_sap.h>
#include <osmocom/sigtran/sccp_helpers.h>
#include <osmocom/sigtran/protocol/sua.h>
#include <osmocom/sigtran/protocol/m3ua.h>

#include <osmocom/gsm/protocol/gsm_08_08.h>
#include <osmocom/gsm/gsm0808.h>
#include <osmocom/gsm/gsm0808_utils.h>


#define SEQUENCE_TIMING	1
#define INITIAL_WAIT SEQUENCE_TIMING * 10
struct osmo_timer_list sequence_timer;


/*
 * FIXME: This is stuff we need to handle the SCCP/M3UA/SCTP connection,
 * this needs to be packed somwhere. By having it global now, we also
 * only support one connection (which is ok for tests)
 */

#define SSN_BSSAP	254 /* SCCP_SSN_BSSAP */
#define SENDER_PC	1
#define RECEIVER_PC	23

struct osmo_sccp_user *scu = NULL;
struct osmo_sccp_instance *sccp = NULL;

/* Sender address for UNITDATA */
static struct osmo_sccp_addr g_calling_addr = {
	.presence = OSMO_SCCP_ADDR_T_SSN | OSMO_SCCP_ADDR_T_PC,
	.ssn = SSN_BSSAP,
	.ri = OSMO_SCCP_RI_SSN_PC,
	.pc = SENDER_PC,
};

/* Receiver address für UNITDATA */
static struct osmo_sccp_addr g_called_addr = {
	.presence = OSMO_SCCP_ADDR_T_SSN | OSMO_SCCP_ADDR_T_PC,
	.ssn = SSN_BSSAP,
	.ri = OSMO_SCCP_RI_SSN_PC,
	.pc = RECEIVER_PC,
};





/*
 * Logging sutff
 */

/* Logging */
static const struct log_info_cat log_info_cat[] = {
};

/* Logging */
static const struct log_info log_info = {
	.cat = log_info_cat,
	.num_cat = ARRAY_SIZE(log_info_cat),
};

/* Initalize logging */
static void init_logging(void)
{
	const int log_cats[] = { DLSS7, DLSUA, DLM3UA, DLSCCP, DLINP };
	unsigned int i;

	osmo_init_logging(&log_info);

	for (i = 0; i < ARRAY_SIZE(log_cats); i++)
		log_set_category_filter(osmo_stderr_target, log_cats[i], 1, LOGL_DEBUG);
}








/* a simple SCCP User which accepts all connections and echos back all
 * DATA + UNITDATA */
static int echo_prim_cb(struct osmo_prim_hdr *oph, void *_scu)
{
	struct osmo_sccp_user *scu = _scu;
	struct osmo_scu_prim *scu_prim = (struct osmo_scu_prim *) oph;
	const uint8_t *data = msgb_l2(oph->msg);
	unsigned int data_len = msgb_l2len(oph->msg);

	switch (OSMO_PRIM_HDR(&scu_prim->oph)) {
	case OSMO_PRIM(OSMO_SCU_PRIM_N_CONNECT, PRIM_OP_INDICATION):
		printf("%s: Accepting N-CONNECT.ind (local_ref=%u)\n",
			__func__, scu_prim->u.connect.conn_id);
		osmo_sccp_tx_conn_resp(scu, scu_prim->u.connect.conn_id,
				     &scu_prim->u.connect.called_addr,
				     data, data_len);
		break;
	case OSMO_PRIM(OSMO_SCU_PRIM_N_DATA, PRIM_OP_INDICATION):
		printf("%s: Echoing N-DATA.ind (local_ref=%u)\n",
			__func__, scu_prim->u.data.conn_id);
		osmo_sccp_tx_data(scu, scu_prim->u.data.conn_id,
				  data, data_len);
		break;
	case OSMO_PRIM(OSMO_SCU_PRIM_N_UNITDATA, PRIM_OP_INDICATION):
		printf("=========================================== GOT SOMETHING! ===========================================\n");
	//	printf("%s: Echoing N-UNITDATA.ind\n", __func__);
	//	osmo_sccp_tx_unitdata(scu, &scu_prim->u.unitdata.called_addr,
	//			      &scu_prim->u.unitdata.calling_addr,
	//			      data, data_len);
		break;
	default:
		printf("%s: Unknown primitive %u:%u\n", __func__,
			oph->primitive, oph->operation);
		break;
	}
	return 0;
}

/* Transmit reset message (UNITDATA) */
void tx_reset()
{
	struct msgb *msg;

	printf("###### |===== RESET =====>|\n");
	msg = gsm0808_create_reset();
	osmo_sccp_tx_unitdata(scu, &g_calling_addr, &g_called_addr, msg->data, msg->len);

	msgb_free(msg);
}

/* Timer callback to update the CMGFX LED-Panel effects */
static void sequence_timer_cb(void *priv)
{
	static unsigned int counter = 0;
	printf("###### Sequence: %i\n", counter);
	fflush(stdout);

	if (!scu) {
		printf("###### Error: No scu!\n");
		goto leave;
	}
		


	if (counter == 0)
		tx_reset();

	counter++;
leave:
	osmo_timer_schedule(&sequence_timer, SEQUENCE_TIMING, 0);
}

int main(void)
{
	/* NOTE: WE PLAY THE BSC RULE HERE:
	 * -- No incoming connections!
	 * -- We never send CLEAR commands! */




	init_logging();
	osmo_ss7_init();


	sccp = osmo_sccp_simple_server(NULL, SENDER_PC, OSMO_SS7_ASP_PROT_M3UA,
					-1, "127.0.0.2");

	/* Why? */
	osmo_sccp_simple_server_add_clnt(sccp, OSMO_SS7_ASP_PROT_M3UA,
					"RECEIVER", RECEIVER_PC, -1, 0, NULL);

	scu = osmo_sccp_user_bind(sccp, "echo", &echo_prim_cb, SSN_BSSAP);
	printf("###### osmo_sccp_user_bind() done!\n");


	sequence_timer.cb = sequence_timer_cb;
	osmo_timer_schedule(&sequence_timer, INITIAL_WAIT, 0);

	printf("###### Enter main loop!\n");
	while (1) {
		osmo_select_main(0);
	}

	return 0;
}
